﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using Emgu.CV;
using Emgu.CV.CvEnum;
using Emgu.CV.Structure;
using Emgu.CV.UI;
//using Emgu.CV.GPU;

namespace TP3_4
{
    public partial class frmMain : Form
    {

        private Capture capture;
        private Image<Bgr, Byte> source;
        private Image<Gray, byte> template;
        private Image<Gray, byte> result;
        private Image<Gray, byte> faceFound;
        private Image<Gray, byte> suspectFace;

        private MCvFont font = new MCvFont(FONT.CV_FONT_HERSHEY_TRIPLEX, 0.5d, 0.5d);
        private HaarCascade face;
        private MCvAvgComp[][] suspectfaceDetected;

        private float maxValue;

        private Boolean canSearchSuspect = false;

        public frmMain()
        {
            InitializeComponent();

            // Nouvelle capture
            capture = new Capture();

            // Chargement du fichier xml permettant la détection de visage
            face = new HaarCascade("haarcascade_frontalface_default.xml");
        }
        
        /**
         * Permet de le redimensionnement des images
         * Cette méthode est utilisée afin d'économiser des ressources
         */
        private float calcRatio(int widthMax, int heightMax, int widthMin, int heightMin)
        {
            float ratio;
            if (widthMax < widthMin && heightMax < heightMin)
                return 1;
            if((float)widthMax / (float)widthMin > (float)heightMax / (float)heightMin)
            {
                ratio = (float)widthMin / (float)widthMax;
            }
            else
            {
                ratio = (float)heightMin / (float)heightMax;
            }

            return ratio;
        }

        /**
         * A chaque tic du timer
         */
        private void timer1_Tick(object sender, EventArgs e)
        {
            try
            {
                // Capture d'une nouvelle image de la caméra de l'ordinateur
                source = capture.QueryFrame();
                // Redimensionnement de l'image en 320 par 240
                float ratio = calcRatio(source.Width, source.Height, 320, 240);
                source = source.Resize((int)Math.Floor(source.Width * ratio), (int)Math.Floor(source.Height * ratio), Emgu.CV.CvEnum.INTER.CV_INTER_CUBIC); ;
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex);
            }

            // Si un suspect a été sélectionné alors on le programme se lance à sa recherche
            if (canSearchSuspect)
            {
                try
                {
                    searchSuspect();
                } catch (Exception ex)
                {
                    Console.WriteLine(ex);
                }
            }

            // Affichage de l'image finale de la caméra
            picCamera.Image = source.ToBitmap();
        }

        /**
         * Méthode permettant la recherche du suspect
         * Cette méthode cherche d'abord les visages dans l'image avant de les comparer à celui du suspect
         */
        private void searchSuspect()
        {
            // Conversion de l'image source en nuances de gris
            Image<Gray, Byte> graySource = source.Convert<Gray, Byte>();

            // Détection des visages
            MCvAvgComp[][] facesDetected = graySource.DetectHaarCascade(
                face,
                1.2,
                10,
                Emgu.CV.CvEnum.HAAR_DETECTION_TYPE.DO_CANNY_PRUNING,
                new Size(20, 20));

            // Chaque visage détecté sera comparé à celui du suspect
            foreach (MCvAvgComp f in facesDetected[0])
            {
                // On récupère le visage détecté
                faceFound = source.Copy(f.rect).Convert<Gray, byte>().Resize(f.rect.Width, f.rect.Width, Emgu.CV.CvEnum.INTER.CV_INTER_CUBIC);

                // Les étapes suivantes sont faites à ce moment car la fonction MatchTemplate nécessite deux images de la même taille. Le visage du suspect doit donc être de la même taille que celui du visage de la caméra
                // Taille que devra avoir le visage du suspect
                float ratio = calcRatio(template.Width, template.Height, f.rect.Width, f.rect.Height);
                int templateWidth = (int)Math.Floor(template.Width * ratio);
                int templateHeight = (int)Math.Floor(template.Height * ratio);

                int i = 0;
                // Recherche du visage du suspect dans l'image
                foreach (MCvAvgComp sf in suspectfaceDetected[0])
                {
                    suspectFace = template.Copy(sf.rect).Convert<Gray, byte>().Resize(templateWidth, templateHeight, Emgu.CV.CvEnum.INTER.CV_INTER_CUBIC);

                    i++;
                    break;
                }
                // Si aucun visage n'est détecté, l'image initiale est conservée pour la comparaison
                if(i==0)
                    suspectFace = template.Convert<Gray, byte>().Resize(templateWidth, templateHeight, Emgu.CV.CvEnum.INTER.CV_INTER_CUBIC);
                
                // Tous les visages détectés sont encadrés en rouge
                source.Draw(f.rect, new Bgr(Color.Red), 2);

                // Comparaison du visage détecté par la caméra avec celui du suspect
                Image<Gray, float> result = faceFound.MatchTemplate(suspectFace, method());
                double[] minValues, maxValues;
                Point[] minLocations, maxLocations;
                // Récupération des seuils de concordances entre les deux images
                result.MinMax(out minValues, out maxValues, out minLocations, out maxLocations);

                // Si la valeur max obtenue est supérieure à celle du seuil, la personne détectée est considérée comme le suspect
                if (maxValues[0] > maxValue)
                {
                    source.Draw("Suspect", ref font, new Point(f.rect.X - 2, f.rect.Y - 2), new Bgr(Color.White));

                }
            }
        }

        /**
         * Choix de la méthode de comparaison pour la fonction MatchTemplate
         */
        private Emgu.CV.CvEnum.TM_TYPE method()
        {
            // Le threshold permet à l'utilisateur d'ajuster le seuil
            float thresold = tkbThreshold.Value / 2;
            // La méthode sélectionnée défini le seuil et renvoie sa valeur pour la fonction MatchTemplate
            if (rdbMethod1.Checked)
            {
                maxValue = 55000000 + thresold * 55000000;
                return Emgu.CV.CvEnum.TM_TYPE.CV_TM_SQDIFF;
            }
            else if (rdbMethod2.Checked)
            {
                maxValue = (float)0.65 + thresold * (float)0.65;
                return Emgu.CV.CvEnum.TM_TYPE.CV_TM_SQDIFF_NORMED;
            }
            else if (rdbMethod3.Checked)
            {
                maxValue = 75000000 + thresold * 75000000;
                return Emgu.CV.CvEnum.TM_TYPE.CV_TM_CCORR;
            }
            else if (rdbMethod4.Checked)
            {
                maxValue = (float)0.75 + thresold * (float)0.75;
                return Emgu.CV.CvEnum.TM_TYPE.CV_TM_CCORR_NORMED;
            }
            else if (rdbMethod5.Checked)
            {
                maxValue = 4000000 + thresold * 4000000;
                return Emgu.CV.CvEnum.TM_TYPE.CV_TM_CCOEFF;
            }
            else {
                maxValue = (float)0.17 + thresold * (float)0.17;
                return Emgu.CV.CvEnum.TM_TYPE.CV_TM_CCOEFF_NORMED;
            }
        }

        /**
         * Ecouteur sur le bouton permettant le choix d'un suspect
         */
        private void btnChangeSuspect_Click(object sender, EventArgs e)
        {
            //Ouverture d'une fenêtre de dialogue permettant la sélection d'un fichier
            openFileDialog.ShowDialog();
        }

        /**
         * Lors de la sélection d'une image par l'utilisateur
         */
        private void openFileDialog_FileOk(object sender, CancelEventArgs e)
        {
            // On récupère l'image du suspect sélectionnée par l'utilisateur
            template = new Image<Gray, byte>(openFileDialog.FileName.ToString());

            // Redimensionnement de l'image du suspect
            float ratio = calcRatio(template.Width, template.Height, 400, 400);
            template = template.Resize((int)Math.Floor(template.Width * ratio), (int)Math.Floor(template.Height * ratio), Emgu.CV.CvEnum.INTER.CV_INTER_CUBIC);

            // Recherche des visages dans l'image du suspect. Cette opération est faite à ce moment pour économiser des ressources pour la suite du programme
            suspectfaceDetected = template.DetectHaarCascade(
                face,
                1.2,
                10,
                Emgu.CV.CvEnum.HAAR_DETECTION_TYPE.DO_CANNY_PRUNING,
                new Size(20, 20));

            picSuspect.Image = template.ToBitmap();           

            // La recherche du suspect peut commencer
            canSearchSuspect = true;
        }
    }
}
